from typing import Dict, List, Optional, Type

from django.db.models import Model, QuerySet
from django.forms import modelform_factory
from django.forms.forms import BaseForm
from django.http import HttpRequest, HttpResponse
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy

from .utils import check_permission


class Action:
    name: str
    verbose_name: str
    permission: Optional[str] = None
    form_class: Optional[Type[BaseForm]] = None

    def __init__(
        self,
        data: Optional[Dict],
        model: Type[Model],
        id: str,
        request: Optional[HttpRequest],
    ):
        self.data = data
        self.is_bound = data is not None

        self.model = model
        self.id = id
        self.data = data
        self.request = request

        self._form: Optional[BaseForm] = None
        self._permission: Optional[str] = None

        assert self.name, "you have to set a name when subclassing action"

    def get_form(self):
        if self._form is None:
            form_class = self.get_form_class()
            if form_class is not None:
                self._form = form_class(data=self.data, prefix=self.id)
        return self._form

    def get_form_class(self) -> Optional[Type[BaseForm]]:
        return self.form_class

    def get_permission(self):
        if self._permission is None and self.permission is not None:
            if isinstance(self.permission, str) and self.model is not None:
                opts = self.model._meta
                self._permission = self.permission.format(
                    action=self, model_name=opts.model_name, app_label=opts.app_label
                )
            else:
                self._permission = self.permission
        return self._permission

    def has_perm(self, user, obj=None):
        # FIXME how should row level permissions be implemented?
        return check_permission(permission=self.get_permission(), user=user, obj=obj)

    def apply(self, queryset: QuerySet) -> Optional[HttpResponse]:
        """
        This applies the action to the queryset.
        Form validation and permission handling happens earlier
        :param queryset:

        :return:
        """
        raise NotImplementedError()

    def get_success_message(self):
        return ""


class DeleteAction(Action):
    name = "delete"
    verbose_name = gettext_lazy("delete")
    permission = "{app_label}.delete_{model_name}"

    def __init__(
        self,
        data: Optional[Dict],
        model: Type[Model],
        id: str,
        request: Optional[HttpRequest],
    ):
        super().__init__(data, model, id, request)
        self.count = 0

    def apply(self, queryset):
        # queryset may be distinct, so we can't use delete directly
        # see https://code.djangoproject.com/ticket/32433
        pks = queryset.values_list("pk", flat=True)
        self.count = self.model._default_manager.filter(pk__in=pks).delete()[0]

    def get_success_message(self):
        return _("Deleted {count} {name}").format(
            count=self.count,
            name=self.model._meta.verbose_name_plural,
        )


class MassUpdateAction(Action):
    name = "update_selected"
    verbose_name = gettext_lazy("update selected")
    permission = "{app_label}.change_{model_name}"

    form_fields: List[str] = []
    form_layout = None
    form_class = None

    def __init__(
        self,
        data: Optional[Dict],
        model: Type[Model],
        id: str,
        request: Optional[HttpRequest],
    ):
        super().__init__(data, model, id, request)
        self.changed = 0

    def get_form_class(self, **kwargs):
        return modelform_factory(model=self.model, fields=self.form_fields, **kwargs)

    def apply(self, queryset):
        form = self.get_form()
        if not form.is_valid():
            raise AssertionError(
                "Ensure that form validation is done before applying this action."
            )

        self.changed = 0
        for instance in queryset:
            instance_changed = False
            for field in form.changed_data:
                current_value = getattr(instance, field, None)
                new_value = form.cleaned_data[field]
                if current_value != new_value:
                    setattr(instance, field, new_value)
                    instance_changed = True
            if instance_changed:
                self.changed += 1
                instance.save()

    def get_success_message(self):
        return _("Updated {count} {name}").format(
            count=self.changed, name=self.model._meta.verbose_name_plural
        )



from django.forms import SelectMultiple
from django.utils.translation import gettext_lazy as _


class BootstrapSelectMultiple(SelectMultiple):
    def __init__(self, attrs=None, choices=()):
        attrs = {} if attrs is None else attrs.copy()
        attrs["class"] = "selectpicker"
        attrs["multiple"] = True
        attrs["data-live-search"] = "true"
        attrs["data-live-search-normalize"] = "true"
        attrs["data-selected-text-format"] = "count > 3"
        attrs["data-actions-box"] = "all"
        attrs["data-deselect-all-text"] = _("Deselect all")
        attrs["data-select-all-text"] = _("Select all")
        attrs["data-count-selected-text"] = _("{0} selected")
        attrs["data-none-selected-text"] = _("None selected")
        super().__init__(attrs, choices)

    class Media:
        js = ["beam/js/bootstrap-select.min.js"]
        css = {
            "all": ["beam/css/bootstrap-select.min.css"],
        }



from typing import List

from dal import autocomplete
from django.db.models import Q
from django.utils.translation import gettext_lazy as _

from beam.components import Component
from beam.urls import UrlKwargDict
from beam.views import ComponentMixin
from beam.viewsets import BaseViewSet


class BaseAutocomplete(ComponentMixin, autocomplete.Select2QuerySetView):
    @property
    def lookup_type(self):
        return self.component.autocomplete_lookup_type

    @property
    def search_fields(self):
        return self.component.autocomplete_search_fields

    def get_result_label(self, item):
        formatter = self.component.autocomplete_result_label
        if formatter:
            return formatter(item)
        return super().get_result_label(item)

    @property
    def model(self):
        return self.component.model

    def get_queryset(self):
        qs = self.component.queryset
        return self.filter_words(self.q, qs)

    def filter_words(self, q, qs):
        assert self.search_fields

        if not q:
            return qs

        if self.lookup_type in ("contains", "icontains"):
            words = self.q.split(" ")
        else:
            words = [self.q]

        qs_filter = Q()
        for word in words:
            word = word.strip()
            if not word:
                continue

            q = Q()
            for field in self.search_fields:
                q |= Q(**{"{}__{}".format(field, self.lookup_type): word})

            qs_filter &= q

        return qs.filter(qs_filter)


class AutocompleteComponent(Component):
    def __init__(
        self,
        autocomplete_search_fields=None,
        autocomplete_result_label=None,
        autocomplete_lookup_type="istartswith",
        **kwargs
    ):
        self.autocomplete_search_fields = autocomplete_search_fields
        self.autocomplete_result_label = autocomplete_result_label
        self.autocomplete_lookup_type = autocomplete_lookup_type
        super().__init__(**kwargs)
        self.show_link = False


class AutocompleteMixin(BaseViewSet):
    """
    A viewset mixin that provides a autocomplete url for the model

    Use `autocomplete_search_fields` to specify the fields to be searched.
    Provide a callable `autocomplete_result_label` that maps results to strings if you want
    to change the string representation of the items.

    """

    autocomplete_view_class = BaseAutocomplete
    autocomplete_url = "autocomplete/"
    autocomplete_url_kwargs: UrlKwargDict = {}
    autocomplete_url_name = None
    autocomplete_verbose_name = _("autocomplete")

    autocomplete_lookup_type = "istartswith"
    autocomplete_search_fields: List[str] = []
    autocomplete_result_label = None
    autocomplete_permission = "{app_label}.view_{model_name}"

    def get_component_classes(self):
        return [
            ("autocomplete", AutocompleteComponent)
        ] + super().get_component_classes()



from typing import Dict, List

from beam.components import BaseComponent


class HTML:
    is_html = True

    def __init__(self, content):
        self.content = content


class VirtualField:
    is_virtual = True

    def __init__(self, name, callback, verbose_name=None):
        self.name = name
        self.callback = callback
        self.verbose_name = verbose_name if verbose_name is not None else str(name)

    def get_value(self, obj=None):
        return self.callback(obj)

    def __str__(self):
        return self.name


def layout_links(
    links: Dict[str, BaseComponent], link_layout: List[str]
) -> List[BaseComponent]:
    if not links:
        return []

    hidden_names = {name[1:] for name in link_layout if name.startswith("!")}
    other_names = [
        name for name in links if name not in link_layout and name not in hidden_names
    ]

    laid_out = []
    for name in link_layout:
        if name == "...":
            for other_name in other_names:
                if links[other_name].show_link:
                    laid_out.append(links[other_name])
        elif name in links and name not in hidden_names and links[name].show_link:
            laid_out.append(links[name])

    return laid_out



from logging import getLogger
from typing import Callable, Dict, Optional, Union

from django.db.models import Model
from django.http import HttpRequest

logger = getLogger(__name__)


undefined = (
    object()
)  # sentinel value for attributes not defined or overwritten on the viewset


UrlKwargGetter = Callable[[Optional[Model], Optional[HttpRequest]], Optional[str]]
"""
A callable that takes an instance and a request and may return a string
as the value for a url kwarg.
"""


UrlKwargDict = Dict[str, Union[str, UrlKwargGetter]]
"""
A dict used to specify url kwargs
"""


def request_kwarg(name: str) -> UrlKwargGetter:
    """
    Return a getter for use in an url_kwargs dict that retrieves a kwarg
    from the request resolver match.
    """

    def getter(obj=None, request=None):
        if request and request.resolver_match and request.resolver_match.kwargs:
            return request.resolver_match.kwargs.get(name, None)
        return None

    return getter



from typing import Optional, Tuple

from django.urls import NoReverseMatch


def check_permission(permission, user, obj):
    if permission is None:
        return True
    if not user:
        return False
    if callable(permission):
        return permission(user, obj=obj)
    # the ModelBackend returns False as soon as we supply an obj so we can't pass that here
    return user.has_perm(permission)


def navigation_component_entry(
    component=None, user=None, request=None
) -> Optional[Tuple[str, str]]:
    """
    Get an optional tuple (label, url) for a given compoment to use in render_navigation
    """
    if not component:
        return None

    if not component.has_perm(user=user, obj=None, request=request):
        return None

    try:
        url = component.reverse(obj=None, request=request)
    except NoReverseMatch:
        return None

    model = getattr(component, "model", None)
    if model and component.name == "list":
        label = model._meta.verbose_name_plural
    else:
        label = component.verbose_name

    return label, url


def reverse_component(component, obj, request, override_kwargs):
    """
    Reverse a component and raise helpful error messages if reversing fails.
    """
    try:
        return component.reverse(
            obj=obj, request=request, override_kwargs=override_kwargs
        )
    except NoReverseMatch as e:
        raise NoReverseMatch(f"Unable to reverse url to {component}: {e}") from e
    except BaseException as e:
        raise Exception(f"Unable to reverse url to {component}: {e}") from e



import decimal
from typing import Dict, List, Sequence, Tuple
from urllib.parse import ParseResult, parse_qsl, urlparse

from django import template
from django.apps import apps
from django.core.exceptions import FieldDoesNotExist
from django.db.models import Manager, Model, QuerySet
from django.db.models.fields.files import FieldFile, ImageFieldFile
from django.db.models.fields.reverse_related import ForeignObjectRel
from django.template import RequestContext
from django.template.loader import get_template
from django.utils.http import urlencode

from beam.components import BaseComponent
from beam.layouts import layout_links
from beam.registry import default_registry, get_viewset_for_model
from beam.utils import navigation_component_entry, reverse_component

register = template.Library()


@register.simple_tag(takes_context=True)
def get_link_url(
    context: RequestContext, component: BaseComponent, obj=None, **override_kwargs
):
    request = getattr(context, "request", None)
    if not component:
        return None

    return reverse_component(
        component, obj=obj, request=request, override_kwargs=override_kwargs
    )


@register.simple_tag
def get_attribute(obj, field):
    if getattr(field, "is_virtual", False):
        return field.get_value(obj)

    if not hasattr(obj, field):
        return None

    if hasattr(obj, "get_{}_display".format(field)):
        return getattr(obj, "get_{}_display".format(field))

    value = getattr(obj, field)

    if isinstance(value, Manager):
        return value.all()

    return value


@register.simple_tag
def get_form_field(form, field):
    try:
        return form[field]
    except KeyError:
        return None


@register.filter
def is_queryset(value):
    return isinstance(value, QuerySet)


@register.simple_tag(takes_context=True)
def get_url_for_related(context, instance, component_name, **override_kwargs):
    if not instance:
        return None

    opts = get_options(instance)

    viewset = context.get("viewset", None)
    if viewset:
        registry = viewset.registry
    else:
        registry = default_registry

    try:
        viewset = get_viewset_for_model(registry, opts.model)
    except KeyError:
        return None

    components = viewset().components
    if component_name not in components:
        return None

    request = getattr(context, "request", None)

    component = components[component_name]

    if request and not component.has_perm(
        request.user, obj=instance, request=request, override_kwargs=override_kwargs
    ):
        return None

    return reverse_component(
        component=component,
        obj=instance,
        request=request,
        override_kwargs=override_kwargs,
    )


@register.filter
def is_image(value):
    return isinstance(value, ImageFieldFile)


@register.filter
def is_file(value):
    return isinstance(value, FieldFile)


@register.filter
def is_bool(value):
    return isinstance(value, bool)


@register.filter
def is_int(value):
    return isinstance(value, int)


@register.filter
def is_float(value):
    return isinstance(value, float)


@register.filter
def is_decimal(value):
    return isinstance(value, decimal.Decimal)


@register.simple_tag
def get_options(instance_or_model):
    """
    Return the _meta options for a given model or instance.
    """
    if not instance_or_model:
        return None

    if isinstance(instance_or_model, Model):
        model = instance_or_model.__class__
    else:
        model = instance_or_model

    return model._meta


@register.filter
def field_verbose_name(instance, field):
    if hasattr(field, "verbose_name"):
        return field.verbose_name

    options = get_options(instance)

    try:
        model_field = options.get_field(field)
    except FieldDoesNotExist:
        model_field = None

    if hasattr(model_field, "verbose_name"):
        return model_field.verbose_name

    if isinstance(model_field, ForeignObjectRel):
        if model_field.multiple:
            return model_field.related_model._meta.verbose_name_plural
        else:
            return model_field.related_model._meta.verbose_name

    value = getattr(instance, field, None)
    return getattr(value, "verbose_name", field.replace("_", " "))


@register.filter(name="any")
def any_filter(iterable):
    if not iterable:
        return False
    return any(iterable)


@register.simple_tag(takes_context=True)
def render_navigation(context):
    request = context.get("request", None)
    user = request.user if request else None

    grouped = []
    for app_label, viewsets_dict in default_registry.items():
        entries = []
        for viewset in viewsets_dict.values():
            entry = navigation_component_entry(
                viewset().links.get("list"), user=user, request=request
            )
            if entry:
                entries.append(entry)

        if not entries:
            continue

        group = {
            "app_label": app_label,
            "app_config": apps.get_app_config(app_label),
            "entries": entries,
        }
        grouped.append(group)
    navigation_template = get_template("beam/partials/navigation.html")
    request = context.get("request", None)
    return navigation_template.render({"apps": grouped, "request": request})


@register.simple_tag()
def fields_to_layout(fields):
    return [[fields]]


@register.simple_tag(takes_context=True)
def sort_link(context, field: str, sorted_fields: Sequence[str]):
    sort_param = context["view"].sort_param
    sort_separator = context["view"].sort_separator

    sorted_fields = list(sorted_fields)
    negated_field = "-" + field

    if field in sorted_fields:
        sorted_fields.remove(field)
        sorted_fields.append(negated_field)
    elif negated_field in sorted_fields:
        sorted_fields.remove(negated_field)
        sorted_fields.append(field)
    else:
        sorted_fields.append(field)

    return preserve_query_string(
        context, **{"page": "", sort_param: sort_separator.join(sorted_fields)}
    )


@register.simple_tag(takes_context=True)
def page_link(context, page_query_string, page_number):
    return preserve_query_string(context, **{page_query_string: page_number})


@register.simple_tag(takes_context=True)
def preserve_query_string(context, **kwargs):
    if "request" not in context:
        raise Exception(
            "The query_string tag requires django.core.context_processors.request"
        )
    request = context["request"]

    get = request.GET.copy()

    for item, value in kwargs.items():
        if value == "":
            get.pop(item, None)
        else:
            get[item] = value

    return "?{}".format(get.urlencode())


def _add_params_to_url_if_new(url, default_params):
    """
    Add params from default_params unless they already exist.
    """
    parsed_url = urlparse(url)
    existing_params = parse_qsl(parsed_url.query)

    params = {}
    params.update(default_params)
    params.update(existing_params)

    return ParseResult(
        parsed_url.scheme,
        parsed_url.netloc,
        parsed_url.path,
        parsed_url.params,
        urlencode(params),
        parsed_url.fragment,
    ).geturl()


PRESERVED_GET_PARAMS = [
    "_popup",
]  # get parameters that should be preserved while following links


@register.simple_tag(takes_context=True)
def get_visible_links(
    context: RequestContext,
    user,
    links: Dict[str, BaseComponent],
    link_layout: List[str],
    obj: Model = None,
    **override_kwargs,
) -> List[Tuple[BaseComponent, str]]:

    get_params = {}
    request = getattr(context, "request", None)
    if request:
        for param in PRESERVED_GET_PARAMS:
            value = request.GET.get(param)
            if value:
                get_params[param] = value

    visible_links = []
    for component in layout_links(links, link_layout):
        if not component.has_perm(
            user, obj=obj, request=request, override_kwargs=override_kwargs
        ):
            continue

        url = reverse_component(
            component=component,
            obj=obj,
            request=request,
            override_kwargs=override_kwargs,
        )

        if url and get_params:
            url = _add_params_to_url_if_new(url, get_params)

        if url:
            visible_links.append((component, url))
    return visible_links



import warnings
from collections import OrderedDict
from typing import TYPE_CHECKING, Dict, Type

if TYPE_CHECKING:
    from .viewsets import BaseViewSet
else:
    BaseViewSet = type


RegistryType = Dict[str, Dict[str, Type[BaseViewSet]]]

test: RegistryType = {"foo": {"bar": BaseViewSet}}


def register(registry: RegistryType, viewset: Type[BaseViewSet]):
    model = viewset.model
    app_label = model._meta.app_label
    model_name = model._meta.model_name

    app_registry = registry.setdefault(app_label, {})
    if model_name in app_registry:
        warnings.warn(
            "Duplicate registration of {} for app label {} with model {}, "
            "beam.viewsets.unregister the existing viewset first.".format(
                viewset, app_label, registry
            )
        )
        return
    app_registry[model_name] = viewset


def unregister(registry, model):
    app_label = model._meta.app_label
    model_name = model._meta.model_name

    if app_label not in registry:
        return

    app_registry = registry[app_label]

    if model_name not in app_registry:
        return

    app_registry.pop(model_name)

    if not app_registry:
        registry.pop(app_label)


def get_viewset_for_model(registry, model):
    app_label = model._meta.app_label
    model_name = model._meta.model_name
    return registry[app_label][model_name]


class ViewsetMetaClass(type):
    def __new__(mcs, name, bases, attrs):
        # Collect components from current class.
        current_component_classes = []
        for key, value in list(attrs.items()):
            if key.endswith("_component"):
                current_component_classes.append((key[: -len("_component")], value))
        attrs["_declared_component_classes"] = OrderedDict(current_component_classes)

        new_class: Type[BaseViewSet] = super().__new__(mcs, name, bases, attrs)

        # Walk through the MRO.
        declared_component_classes = OrderedDict()
        for base in reversed(new_class.__mro__):
            # Collect fields from base class.
            if hasattr(base, "_declared_component_classes"):
                declared_component_classes.update(base._declared_component_classes)

            # Field shadowing.
            for attr, value in base.__dict__.items():
                if value is None and attr in declared_component_classes:
                    declared_component_classes.pop(attr)

        new_class._component_classes = list(declared_component_classes.items())
        new_class._declared_component_classes = declared_component_classes

        if (
            new_class.registry is not None
            and getattr(new_class, "model", None) is not None
        ):
            register(new_class.registry, new_class)

        return new_class


default_registry: RegistryType = {}



from django.urls import include, path

from beam.contrib.auth.views import GroupViewSet, UserViewSet

urlpatterns = [
    path("user/", include(UserViewSet().get_urls())),
    path("group/", include(GroupViewSet().get_urls())),
]



from django.contrib.auth import get_user_model
from django.contrib.auth.forms import UserCreationForm
from django.contrib.auth.models import Group, Permission
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

import beam.inlines
from beam import ViewSet
from beam.contrib.auth.forms import GroupForm, UserUpdateForm
from beam.layouts import VirtualField


class GroupInline(beam.inlines.RelatedInline):
    model = Group
    foreign_key_field = "nope"
    layout = [
        [
            [
                VirtualField(
                    "group",
                    callback=lambda obj: obj,
                    verbose_name=Group._meta.get_field("name").verbose_name,
                )
            ]
        ]
    ]

    def get_prefix(self):
        return "groups"

    def get_queryset(self):
        return Group.objects.filter(user=self.parent_instance)


class UserPermissionInline(beam.inlines.TabularRelatedInline):
    model = Permission
    foreign_key_field = "nope"
    fields = ["name", "content_type", "codename"]

    def get_prefix(self):
        return "permissions"

    def get_queryset(self):
        return Permission.objects.filter(user=self.parent_instance).select_related(
            "content_type"
        )


class UserViewSet(ViewSet):
    model = get_user_model()

    detail_inline_classes = [UserPermissionInline, GroupInline]

    detail_layout = [
        [
            [
                "username",
                "first_name",
                "last_name",
                "email",
                "is_active",
                "is_staff",
                "is_superuser",
                VirtualField(
                    "dates_header",
                    lambda obj: "",
                    verbose_name=mark_safe(
                        "<h4 class='mt-2'>{}</h4>".format(_("Important dates"))
                    ),
                ),
                "last_login",
                "date_joined",
            ]
        ]
    ]
    update_layout = [
        [["username", "email"]],
        [["first_name", "last_name"]],
        [["is_active", "is_staff", "is_superuser"]],
        [["groups"]],
        [["user_permissions"]],
        [["last_login", "date_joined"]],
    ]
    list_fields = [
        "username",
        "email",
        "first_name",
        "last_name",
        "is_staff",
    ]
    list_sort_fields = list_fields
    list_search_fields = ["username", "email", "first_name", "last_name"]
    list_filterset_fields = ["is_active", "is_staff", "is_superuser", "groups"]

    create_form_class = UserCreationForm
    update_form_class = UserUpdateForm


class UserInline(beam.inlines.RelatedInline):
    model = get_user_model()
    foreign_key_field = "nope"
    layout = [
        [
            [
                VirtualField(
                    "user",
                    callback=lambda obj: obj,
                    verbose_name=get_user_model()._meta.verbose_name,
                )
            ]
        ]
    ]

    def get_prefix(self):
        return "users"

    def get_queryset(self):
        return get_user_model().objects.filter(groups=self.parent_instance)


class GroupPermissionInline(beam.inlines.TabularRelatedInline):
    model = Permission
    foreign_key_field = "nope"
    fields = ["name", "content_type", "codename"]

    def get_prefix(self):
        return "permissions"

    def get_queryset(self):
        return Permission.objects.filter(group=self.parent_instance).select_related(
            "content_type"
        )


class GroupViewSet(ViewSet):
    model = Group
    list_fields = ["name"]
    fields = ["name", "permissions"]
    list_search_fields = ["name"]
    list_sort_fields = ["name"]
    detail_inline_classes = [GroupPermissionInline, UserInline]
    detail_fields = [
        "name",
    ]
    form_class = GroupForm



from django import forms
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group

from beam.themes.bootstrap4.widgets import BootstrapSelectMultiple


class UserUpdateForm(forms.ModelForm):
    class Meta:
        model = get_user_model()
        fields = [
            "username",
            "first_name",
            "last_name",
            "email",
            "is_active",
            "is_staff",
            "is_superuser",
            "groups",
            "user_permissions",
        ]
        widgets = {
            "groups": BootstrapSelectMultiple(),
            "user_permissions": BootstrapSelectMultiple(),
        }


class GroupForm(forms.ModelForm):
    class Meta:
        model = Group
        fields = [
            "name",
            "permissions",
        ]
        widgets = {
            "permissions": BootstrapSelectMultiple(),
        }
